using System;
using System.Collections.Generic;
using System.Text;

using JetBrains.ProjectModel;
using JetBrains.ReSharper;
using JetBrains.ReSharper.Daemon;
using JetBrains.ReSharper.Psi;
using JetBrains.ReSharper.Psi.CSharp;
using JetBrains.ReSharper.Psi.CSharp.Impl.Tree;
using JetBrains.ReSharper.Psi.CSharp.Tree;
using JetBrains.ReSharper.Psi.ExtensionsAPI.Tree;
using JetBrains.ReSharper.Psi.Tree;
using JetBrains.ReSharper.TextControl;
using JetBrains.Shell;
using JetBrains.Util;

namespace BrevanHoward.ResharperPlugins.MakeMethodThreadSafePlugin
{
    [ContextAction(Description = "Makes a method or property threadsafe", Name = "Make threadsafe", Priority = -1,
        Group = "C#")]
    public class MakeMethodThreadSafeAction: IContextAction, IBulbItem
    {
        public MakeMethodThreadSafeAction(ISolution solution, ITextControl textControl)
        {
            _solution = solution;
            _textControl = textControl;
        }

        private readonly ISolution _solution;
        private readonly ITextControl _textControl;

        private CSharpElementFactory Factory
        {
            get {return CSharpElementFactory.GetInstance(_solution);}
        }

        private void DoActualWork(IMethodDeclaration methodDeclaration)
        {
            CSharpElementFactory factory = Factory;
            string delegateName = methodDeclaration.DeclaredName + "Delegate";
            string delegateTemplate = string.Format("private delegate $0 {0}($1);", delegateName);
            IElement element = methodDeclaration.DeclaredElement as IElement;
            if (element == null)
                return;
            IMethod method = (IMethod)element;
            IType returnType = method.ReturnType;

            IDelegateDeclaration delegateDeclaration =
                factory.CreateTypeMemberDeclaration(delegateTemplate, returnType, GetParamList(methodDeclaration)) as
                IDelegateDeclaration;
            if (delegateDeclaration == null)
                return;

            IClassDeclaration classDeclaration = methodDeclaration.GetContainingTypeDeclaration() as IClassDeclaration;
            if (classDeclaration == null)
                return;

            classDeclaration.AddClassMemberDeclarationBefore(delegateDeclaration, methodDeclaration);

            classDeclaration.MemberDeclarations.Add(delegateDeclaration);

            string invocation =
                CreateDelegateInvocation(
                    delegateName, methodDeclaration.DeclaredName, methodDeclaration.ParameterDeclarations);

            IStatement conditionalStatement;
            string longPresentableName = returnType.GetLongPresentableName(CSharpLanguageService.CSHARP);
            if (longPresentableName == "void")
                conditionalStatement = factory.CreateStatement("if (InvokeRequired) {$0; return;}", invocation);
            else
                conditionalStatement =
                    factory.CreateStatement("if (InvokeRequired) return ($0) ($1);", longPresentableName, invocation);

            methodDeclaration.Body.AddStatementBefore(conditionalStatement, methodDeclaration.Body.Statements[0]);
        }

        private string CreateDelegateInvocation(
            string delegateName, string declaredName, IEnumerable<IRegularParameterDeclaration> parameterDeclarations)
        {
            StringBuilder sb = new StringBuilder();
            sb.AppendFormat("this.Invoke(new {0}({1})", delegateName, declaredName);
            foreach (IRegularParameterDeclaration parameterDeclaration in parameterDeclarations)
                sb.Append(',').Append(parameterDeclaration.DeclaredName);
            sb.Append(')');
            return sb.ToString();
        }

        private IMethodDeclaration GetMethodDeclaration(IElement element)
        {
            if (!IsElementKeywordPublic(element as CSharpToken))
                return null;

            IModifiersListNode modifiersListNode = element.GetContainingElement<IModifiersListNode>(false);
            if (modifiersListNode == null)
                return null;

            IMethodDeclaration methodDeclaration = modifiersListNode.Parent as IMethodDeclaration;
            if (methodDeclaration == null)
                return null;

            if (!HasThreadSafeInfrastructure(methodDeclaration.GetContainingTypeDeclaration() as IClassDeclaration))
                return null;
            return methodDeclaration;
        }

        private string GetParamList(IDeclarationWithParameters methodDeclaration)
        {
            IList<IRegularParameterDeclaration> parameterDeclarations = methodDeclaration.ParameterDeclarations;
            List<string> paramList = new List<string>();
            for (int index = 0; index < parameterDeclarations.Count; index++)
                paramList.Add(parameterDeclarations[index].GetText());
            return string.Join(",", paramList.ToArray());
        }

        private bool HasMethod(IList<IMethodDeclaration> declarations, string name)
        {
            for (int index = 0; index < declarations.Count; index++)
                if (declarations[index].DeclaredName == name)
                    return true;
            return false;
        }

        private bool HasMethod(IMethod[] methods, string name)
        {
            for (int index = 0; index < methods.Length; index++)
                if (methods[index].ShortName == name)
                    return true;
            return false;
        }

        private bool HasProperty(IList<IPropertyDeclaration> declarations, string name)
        {
            for (int index = 0; index < declarations.Count; index++)
                if (declarations[index].DeclaredName == name)
                    return true;
            return false;
        }

        private bool HasProperty(IProperty[] properties, string name)
        {
            for (int index = 0; index < properties.Length; index++)
                if (properties[index].ShortName == name)
                    return true;
            return false;
        }

        private bool HasThreadSafeInfrastructure(IClassLikeDeclaration cd)
        {
            if (cd == null)
                return false;
            if (HasProperty(cd.PropertyDeclarations, "InvokeRequired") && HasMethod(cd.MethodDeclarations, "Invoke"))
                return true;

            IList<IDeclaredType> superTypes = cd.SuperTypes;
            foreach (IDeclaredType superType in superTypes)
                if (HasThreadSafeInfrastructure(superType))
                    return true;

            return false;
        }

        private bool HasThreadSafeInfrastructure(IDeclaredType superType)
        {
            ITypeElement te = superType.GetTypeElement();
            if (te == null)
                return false;
            IMethod[] methods = te.Methods;
            IProperty[] properties = te.Properties;
            if (HasProperty(properties, "InvokeRequired") && HasMethod(methods, "Invoke"))
                return true;
            for (int index = 0; index < superType.GetSuperTypes().Length; index++)
                if (HasThreadSafeInfrastructure(superType.GetSuperTypes()[index]))
                    return true;
            return false;
        }

        private bool IsElementKeywordPublic(TreeElement cSharpToken)
        {
            if (cSharpToken == null)
                return false;
            if (!cSharpToken.IsPhysical())
                return false;
            if (!cSharpToken.IsValid())
                return false;
            if (cSharpToken.type.ToString() != "PUBLIC_KEYWORD" || cSharpToken.GetText() != "public")
                return false;
            return true;
        }

        IBulbItem[] IBulbAction.Items
        {
            get {return new IBulbItem[] {this};}
        }

        string IBulbItem.Text
        {
            get {return "Make method or property threadsafe";}
        }

        public void Execute(ISolution solution, ITextControl textControl)
        {
            IMethodDeclaration methodDeclaration =
                GetMethodDeclaration(TextControlToPsi.GetSourceTokenAtCaret(_solution, _textControl));

            if (methodDeclaration == null)
                return;

            using (CommandCookie.Create("MakeThreadSafe"))
            {
                if (methodDeclaration.Body.Statements.Count == 0)
                    return;

                PsiManager.GetInstance(_solution).DoTransaction(delegate {DoActualWork(methodDeclaration);});
            }
        }

        bool IBulbAction.IsAvailable(IUserDataHolder cache)
        {
            return GetMethodDeclaration(cache.GetData(ContextActionsDataBuilder.ELEMENT_KEY)) != null;
        }
    }

    /// <summary>We need such a complex construction because IContextAction.IsAvailable 
    /// gets called on a dedicated thread, and we can't access caret position on that 
    /// thread. So we cache necessary data in the IContextActionDataBuilder.Build method 
    /// and use them in IContextAction.IsAvailable later.</summary>
    [ContextActionDataBuilder]
    internal class ContextActionsDataBuilder: IContextActionDataBuilder
    {
        public static readonly Key<IElement> ELEMENT_KEY = new Key<IElement>("Element");

        public void Build(ISolution solution, ITextControl textControl, IUserDataHolder cache)
        {
            IElement element = TextControlToPsi.GetSourceTokenAtCaret(solution, textControl);
            cache.PutData(ELEMENT_KEY, element);
        }
    }
}